/* -*- c++ -*- */
/* 
 * Copyright 2015 <+YOU OR YOUR COMPANY+>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "ngham_encode_impl.h"
#include <gnuradio/fec/rs.h>
#include <gnuradio/fec/api.h>
#include <stdio.h>


namespace gr {
  namespace ngham {

    ngham_encode::sptr
    ngham_encode::make()
    {
      return gnuradio::get_initial_sptr
        (new ngham_encode_impl());
    }

    /*
     * The private constructor
     */
    ngham_encode_impl::ngham_encode_impl()
      : gr::block("ngham_encode",
              gr::io_signature::make(1, 1, sizeof(char)),
              gr::io_signature::make(1, 1, sizeof(char))) //,
			  /*size_nr(0), d_len(0), NGH_CODEWORD_START(NGH_PREAMBLE_SIZE+NGH_SYNC_SIZE+NGH_SIZE_TAG_SIZE), ngham_flags(0)*/
    {
    	//if ((pl_len == 0) || (pl_len > NGH_PL_SIZE[NGH_SIZES-1])) return;
    	printf ("Reached constructor \n");
    }

    /*
     * Our virtual destructor.
     */
    ngham_encode_impl::~ngham_encode_impl()
    {
    }

    void
    ngham_encode_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
    	   	ninput_items_required[0] = noutput_items;
    }
		
    int
    ngham_encode_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
        const unsigned char *in = (const unsigned char *) input_items[0];
        unsigned char *out = (unsigned char *) output_items[0];
        
        pl_len = ninput_items[0];
        d_len = 0;
        out_len = 0;
        size_nr = 0;
        
        // Do <+signal processing+>
        printf ("Reached signal processing part\n");
        printf("%d\n",pl_len);
        // Check size and find control block for smallest possible RS codeword
        while ((pl_len) > NGH_PL_SIZE[size_nr]) size_nr++;
        
        // Insert preamble, sync and size-tag
        for (j = 0; j < NGH_PREAMBLE_SIZE; j++) out[out_len++] = NGH_PREAMBLE;              						// Insert preamble
        for (j = 0; j < NGH_SYNC_SIZE; j++) out[out_len++] = NGH_SYNC[j];                              			    // Insert syncword
        for (j = 0; j < NGH_SIZE_TAG_SIZE; j++) out[out_len++] = (NGH_SIZE_TAG[size_nr] >> (16-j*8)) & 0xff;        // Insert size tag
        
        // Prepare content of codeword
        d[d_len++] = ((NGH_PL_SIZE[size_nr] - (pl_len) & 0x1f) & 0xe0);
       
        for (j=0; j < pl_len; j++) d[d_len++] = in[j];																// Insert data
        crc = crc_ccitt(&d[0], pl_len+1);																			// Insert CRC
        for(j=0; j<2; j++) d[d_len++] = (crc >> (8-j*8)) & 0xff;
        
        for (j = pl_len+3; j<NGH_PL_SIZE_FULL[size_nr]; j++) d[d_len++] = 0;										// Insert padding
        
        // Generate parity data
        encode_rs_char(&rs_cb[size_nr], &d[0], &d[d_len]);
        d_len += NGH_PAR_SIZE[size_nr];
        //for (j=0; j<NGH_PAR_SIZE[size_nr]; j++) d[d_len++] = 0xAA;												// Insert dummy parity bits for testing
        
        // Scramble
        for (j=0; j<NGH_PL_PAR_SIZE[size_nr]; j++) d[j] ^= ccsds_poly[j];
        //for (j=0; j<NGH_PL_PAR_SIZE[size_nr]; j++) d[NGH_CODEWORD_START+j] ^= ccsds_poly[j];
       	
        for(j = 0; j<d_len; j++) out[out_len++] = d[j];
                	
        // Tell runtime system how many input items we consumed on
        // each input stream.
       	consume_each(pl_len);

       	// Tell runtime system how many output items we produced.
        //return noutput_items;
        printf("Signal processing finished!\n");
        return out_len;
    }

  } /* namespace ngham */
} /* namespace gr */

